#!BPY

"""
Name: 'BVH 4 BLENDER'
Blender: 234
Group: 'Wizards'
Tooltip: 'Create Armature on BVH Empties'
"""

######################################################
# BVH4BLEND.py v1.0
#
# Script for generating armature on BVH empties
#
#####################################################
# This script generates an armature and make bones
# follow empties created by Blender BVH import 
# How to use it :
# - Import a bvh in Blender (Fie->Import->BVH)
# - Launch this script (Alt-P)
# - Set up variables : 
#   "startframe"  the first frame of your anim
#   "endframe"  the last frame of your anim
#   "hipbonename" the name of the main bone
# - Press "Create Armature"
######################################################
# Author : Jean-Baptiste PERIN (jb_perin@yahoo.fr)
######################################################



import Blender
from Blender import Mathutils
import math

dicEmptiesRestMatrix= {}
dicBoneRestMatrix= {}

########################################################################
#
#          UTILITY FUNCTIONS FOR HANDLING BONES AND EMPTIES
#
########################################################################

#########
# Cette fonction renvoie la liste des empties
# in  : 
# out : emp_list (List of Object) la liste des objets de type "Empty"
#########
def getEmpties():
	emp_list = []
	objs = Blender.Object.Get()
	for o in objs:
		if o.getType()=="Empty":
			emp_list.append(o)
	return emp_list

#########
# Cette fonction renvoie la liste des empties
# in  : 
# out : emp_list (List of Object) la liste des objets de type "Empty"
#########
def getEmpty(name):
	p = None
	objs = Blender.Object.Get()
	for o in objs:
		if o.getType()=="Empty" and o.getName()==name:
			p = o
	return p

#########
# Cette fonction fournit la liste des enfants d'un empty
# in  : emp (Object) un empty
#       emp_list (List of Object) la liste des empties
# out : children (List of Object) la liste des empties enfants de 'empty'
#########
def getChildren(emp, emp_list):
	children = []
	root_emp = getRootEmpties(emp_list)
	for em in emp_list:
		if (em.getName() != emp.getName()) and (em != root_emp[0]):
			if (em.getParent().getName() == emp.getName()):
				children.append(em)
	return children



#########
# Cette fonction renvoie la liste des empties n'ayant pas de parent
# in  : emp_list (List) une liste d'empties
# out : root (List) le (ou les) empty n'ayant pas de parent
#########
def getRootEmpties(emp_list):
	root = []
	for em in emp_list:
		if em.getParent() == None:
			root.append(em)
	return root


#########
# Cette fonction renvoie le bone de nom 'name' dans l'armature 'armature'
# in  : armature (Armature) l'armature dans laquelle cherchait le bone
#       name (String) le nom de l'os a chercher
# out : p (Bone) 
#########
def getBone(armature, name):
	p = None
	bones = armature.getBones()
	for i in bones:
		if i.getName() == name:
			p = i
			break
	return p




########################################################################
#
# FUNCTIONS FOR COMPUTING POSITION AND ROTATION OF BONES 
#
########################################################################

def computeRootQuat2(empty, bone):
	emprest = dicEmptiesRestMatrix[empty.getName()].rotationPart()
	emprest.transpose()
	invemprest= dicEmptiesRestMatrix[empty.getName()].rotationPart()
	invemprest.transpose()
	invemprest.invert()
	emprot = empty.getMatrix('worldspace').rotationPart()
	emprot.transpose()
	T2=invemprest*emprot
	T2.transpose()
	transglobal =emprest*T2*invemprest
	
	bonerest = dicBoneRestMatrix[bone.getName()].rotationPart()
	bonerest.transpose()
	invbonerest = dicBoneRestMatrix[bone.getName()].rotationPart()
	invbonerest.transpose()
	invbonerest.invert()
	mat = invbonerest*transglobal*bonerest
	
	return (mat.toQuat())




#########
# Cette fonction 
# in  : 
# out : 
#########
def computeRootPos(empty, bone):
	vec = empty.getMatrix('worldspace').translationPart() - dicBoneRestMatrix[bone.getName()].translationPart()
	mat = dicBoneRestMatrix[bone.getName()].rotationPart()
	mat.transpose()
	vec2 =  Mathutils.VecMultMat ( vec, mat)
	return vec2


########################################################################
#
#            FUNCTIONS FOR CREATING AND MOVING ARMATURES
#
########################################################################

#########
# Cette fonction 
# in  : 
# out : 
#########
def createBone (armature, empty, bone, empties):
	children = getChildren(empty, empties)
	for ch in children:
		if len(children) >= 2:
			bonename = empty.getName()[1:len(empty.getName())]+'_'+ch.getName()[1:len(ch.getName())]
		else :
			bonename = empty.getName()[1:len(empty.getName())]
		b=Blender.Armature.Bone.New(bonename)
		b.setHead(empty.getMatrix('worldspace').translationPart())
		b.setTail(ch.getMatrix('worldspace').translationPart())
		#b.setParent(bone)
		dicEmptiesRestMatrix[empty.getName()] = empty.getMatrix('worldspace')
		armature.addBone(b)
		dicBoneRestMatrix[b.getName()] = b.getRestMatrix()

		createBone(armature, ch, b, empties)

	

#########
# Cette fonction fabrique une arma
# in  : 
# out : 
#########
def createArmature (rootEmpty, empties):
	armData = Blender.Armature.New('monArmature')
	children = getChildren(rootEmpty, empties)
	for ch in children:
		b=Blender.Armature.Bone.New(rootEmpty.getName()[1:len(rootEmpty.getName())] + ch.getName()[1:len(ch.getName())])
		b.setHead(rootEmpty.getMatrix('worldspace').translationPart())
		b.setTail(ch.getMatrix('worldspace').translationPart())
		armData.addBone(b)
		dicEmptiesRestMatrix[rootEmpty.getName()] = ch.getMatrix('worldspace')
		dicBoneRestMatrix[b.getName()] = b.getRestMatrix()
		createBone(armData, ch, b, empties)
	return armData

#########
# Cette fonction 
# in  : 
# out : 
#########
def moveBones(armature, empty, empties):
	children = getChildren(empty, empties)
	for ch in children:
		if len(children) >= 2:
			bonename = empty.getName()[1:len(empty.getName())]+'_'+ch.getName()[1:len(ch.getName())]
		else :
			bonename = empty.getName()[1:len(empty.getName())]

		bone = getBone(armature, bonename)
		bone.setLoc(computeRootPos(empty,bone))
		bone.setQuat(computeRootQuat2(empty,bone))
	for ch in children:
		chch = getChildren(ch, empties)
		if len(chch) >= 1:
			moveBones(armature, ch, empties)
		#bo = getBone(armature, ch.getName()[1:len(ch.getName())])
		#if bo != None:
		#	moveBones(armature, ch, empties)
		#else:
		#	bo = getBone(armature, ch.getName()[1:len(ch.getName())]+'_'+ch.getName()[1:len(ch.getName())])


#########
# Cette fonction 
# in  : 
# out : 
#########
def moveArmature (armature, empties):
	root = getRootEmpties(empties)
	children = getChildren(root[0], empties)
	for ch in children:
		b=getBone(armature, root[0].getName()[1:len(root[0].getName())] + ch.getName()[1:len(ch.getName())])
		b.setLoc(computeRootPos(root[0], b))
		b.setQuat(computeRootQuat2(root[0], b))
		moveBones(armature, ch, empties)




########################################################################
#
#                  MAIN PROGRAM
#
########################################################################

def Main():

	global endframe, startframe,hipbonename


	print "*****START*****"

	Blender.Set("curframe",startframe)

	##-----------
	## Positionnement des empties
	##-----------
	em0 = Blender.Object.Get(hipbonename)

	Blender.Redraw()



	##-----------
	## Creation de l'armature et des os
	##-----------


	armData = createArmature(em0, getEmpties())
	#armData.drawAxes(1)
	#armData.drawNames(1)

	##-----------
	## Ajout de l'armature a la scene
	##-----------
	armObj=Blender.Object.New('Armature', 'OBArmature')

	armObj.link(armData)
	scn = Blender.Scene.getCurrent()
	scn.link (armObj)
	Blender.Redraw()

	##-----------
	## Enregistrement de la position  de l'armature
	##-----------

	bones = armData.getBones()
	for bo in bones:
		bo.setPose([Blender.Armature.Bone.ROT, Blender.Armature.Bone.LOC]) 

	Blender.Redraw()


	while endframe >= Blender.Get("curframe"):
		##-----------
		## Passage a la frame suivante
		##-----------
		num_frame = Blender.Get("curframe")
		Blender.Set("curframe", num_frame+1)



		##-----------
		## Positionnement des os
		##-----------

		moveArmature(armData, getEmpties())


		##-----------
		## Enregistrement de la position  de l'armature
		##-----------

		bones = armData.getBones()
		for bo in bones:
			bo.setPose([Blender.Armature.Bone.ROT, Blender.Armature.Bone.LOC]) 

	#Blender.Redraw()

	Blender.Set("curframe",startframe)
	Blender.Redraw()

	print "*****END*****"


########################################################################
#
#            GUI FUNCTIONS AND VARIABLES
#
########################################################################

EFrame = Blender.Draw.Create(0)
SFrame2 = Blender.Draw.Create(0)
HBName = Blender.Draw.Create(0)
Msg = ' '

def event (evt, val):
	if evt == Blender.Draw.ESCKEY:
		Blender.Draw.Exit()
	return

def button_event(evt):
	global EFrame, SFrame2, HBName, Msg
	global endframe, startframe, hipbonename
	if evt==1:
		startframe = eval(SFrame2.val)
		endframe =  eval(EFrame.val)
		hipbonename = HBName.val
		if startframe>=endframe:
			Msg = 'Start frame must be lower than End frame'
		else:
			ob = getEmpty(hipbonename)
			if (ob!=None): 
				if ob.getParent()!=None:
					Msg = 'Empty '+hipbonename+ ' is not a root bone.'
				else:  
					Blender.Draw.Exit()
					Main()
			else:
				Msg = 'Empty '+ hipbonename+ ' not found'
				
		Blender.Draw.Redraw(1)

def GUI():
	global EFrame, SFrame2, HBName, Msg
	Blender.BGL.glClearColor(0,0,1,1)
	Blender.BGL.glClear(Blender.BGL.GL_COLOR_BUFFER_BIT)
	Blender.BGL.glColor3f(1,1,1)
	Blender.BGL.glRasterPos2i(20,150)
	Blender.Draw.Text ("BVH 4 BLENDER v1.0 by Jean-Baptiste PERIN", 'normal')
	EFrame = Blender.Draw.String("Endframe: ", -1, 20, 100, 250, 20, '250', 100)
	HBName = Blender.Draw.String("HipBoneName: ", -1, 20, 75, 250, 20, '_Hips', 100)
	SFrame2 = Blender.Draw.String("Startframe: ", -1, 20, 125, 250, 20, '2', 100)
	Blender.Draw.Toggle("Create Armature", 1, 20, 10, 100, 20, 0, "Create Armature")
	Blender.BGL.glRasterPos2i(20,40)
	Blender.Draw.Text (Msg, 'normal')


Blender.Draw.Register(GUI, event, button_event)
